/* 
	Corsix's Image Retargetr
    Copyright (C) 2007 Peter Cawley

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "IndexedRetarget.h"
#include <stack>

IndexedRetarget::IndexedRetarget()
{
	m_pCurrentImage = 0;
	m_pOriginalImage = 0;
}

IndexedRetarget::~IndexedRetarget()
{
	delete m_pCurrentImage;
}

const unsigned char* IndexedRetarget::GetImageData() const
{
	return m_pCurrentImage->pData;
}

const unsigned char* IndexedRetarget::GetImageEnergyVisualisation()
{
	unsigned char*& pVis = m_pCurrentImage->pParent->pIndexVisualisation;
	if(!pVis)
		m_pCurrentImage->pParent->visualiseIndex();
	return pVis;
}

template <class T> static inline T min (T a, T b) {return a < b ? a : b;}
template <class T> static inline T max (T a, T b) {return a > b ? a : b;}

void IndexedRetarget::_image_t::visualiseIndex()
{
	delete[] pIndexVisualisation;
	pIndexVisualisation = new unsigned char[iPitch * iHeight];
	unsigned char* pOutput = pIndexVisualisation;
	long* pInput = pIndexMap;

	for(int iY = 0; iY < iHeight; ++iY)
	{
		for(int iX = 0; iX < iWidth; ++iX)
		{
			long iVal = *pInput;
			iVal /= 4;

			pOutput[0] = max(min(iVal - 256, (long)255), (long)0);
			pOutput[1] = min(iVal, (long)255);
			pOutput[2] = 0;

			pOutput += 3;
			++pInput;
		}
	}
}

int IndexedRetarget::GetWidth() const
{
	return m_pCurrentImage->iWidth;
}

int IndexedRetarget::GetHeight() const
{
	return m_pCurrentImage->iHeight;
}

bool IndexedRetarget::Init(int iWidth, int iHeight, const unsigned char* pData)
{
	delete m_pCurrentImage;
	m_pCurrentImage = 0;
	m_pOriginalImage = 0;

	m_pOriginalImage = new _image_t;
	m_pOriginalImage->iWidth = iWidth;
	m_pOriginalImage->iPitch = iWidth * 3;
	m_pOriginalImage->iHeight = iHeight;
	m_pOriginalImage->pData = new unsigned char[iWidth * iHeight * 3];

	if(m_pOriginalImage->pData == 0)
		return false;

	if(pData == 0)
		memset(m_pOriginalImage->pData, 0, iWidth * iHeight * 3);
	else
		memcpy(m_pOriginalImage->pData, pData, iWidth * iHeight * 3);

	m_pOriginalImage->CalculateEnergyData();

	if(iWidth > iHeight)
		m_pOriginalImage->CalculateIndexColumns();
	else
		m_pOriginalImage->CalculateIndexRows();

	m_pCurrentImage = new _image_t;
	m_pCurrentImage->pParent = m_pOriginalImage;
	m_pCurrentImage->iWidth = iWidth;
	m_pCurrentImage->iPitch = iWidth * 3;
	m_pCurrentImage->iHeight = iHeight;
	m_pCurrentImage->pData = new unsigned char[iWidth * iHeight * 3];

	if(m_pCurrentImage->pData == 0)
		return false;

	if(pData == 0)
		memset(m_pCurrentImage->pData, 0, iWidth * iHeight * 3);
	else
		memcpy(m_pCurrentImage->pData, pData, iWidth * iHeight * 3);

	return true;
}

IndexedRetarget::_image_t::_image_t()
{
	pParent = 0;
	pIndexMap = 0;
	pIndexVisualisation = 0;
	pData = 0;
	pEnergyData = 0;
	iWidth = 0;
	iPitch = 0;
	iHeight = 0;
}

IndexedRetarget::_image_t::~_image_t()
{
	delete[] pIndexMap;
	delete[] pIndexVisualisation;
	delete[] pData;
	delete[] pEnergyData;
	delete pParent;
}

long IndexedRetarget::_image_t::_pathfind_t::getHCost()
{
	if(!bGotHCost)
	{
		bGotHCost = true;
		//iHCost = getCost(); /*
		iHCost = *pCostSrc;
		if(pChild[1])
			iHCost += pChild[1]->getHCost();
		//*/
		
	}
	return iHCost;
}

long IndexedRetarget::_image_t::_pathfind_t::getCost()
{
	if(pShortestChild)
		return iCost;
	else
	{
		iCost = *pCostSrc;

		if(pChild[1] == 0)
		{
		}
		else if(pChild[0] && pChild[2])
		{
			int iChild0 = pChild[0]->getHCost();
			int iChild1 = pChild[1]->getHCost();
			int iChild2 = pChild[2]->getHCost();

			if(iChild0 <= iChild1 && iChild0 <= iChild2)
			{
				iCost += pChild[0]->getCost();
				pShortestChild = pChild[0];
			}
			else if(iChild1 <= iChild0 && iChild1 <= iChild2)
			{
				iCost += pChild[1]->getCost();;
				pShortestChild = pChild[1];
			}
			else
			{
				iCost += pChild[2]->getCost();
				pShortestChild = pChild[2];
			}
		}
		else if(pChild[0])
		{
			int iChild0 = pChild[0]->getHCost();
			int iChild1 = pChild[1]->getHCost();
			if(iChild0 <= iChild1)
			{
				iCost += pChild[0]->getCost();;
				pShortestChild = pChild[0];
			}
			else
			{
				iCost += pChild[1]->getCost();;
				pShortestChild = pChild[1];
			}
		}
		else if(pChild[2])
		{
			int iChild1 = pChild[1]->getHCost();
			int iChild2 = pChild[2]->getHCost();
			if(iChild1 <= iChild2)
			{
				iCost += pChild[1]->getCost();;
				pShortestChild = pChild[1];
			}
			else
			{
				iCost += pChild[2]->getCost();;
				pShortestChild = pChild[2];
			}
		}
		else
		{
			iCost += pChild[1]->getCost();
			pShortestChild = pChild[1];
		}
	}
	return iCost;
}

void IndexedRetarget::_image_t::CalculateEnergyData()
{
	delete[] pEnergyData;
	pEnergyData = new long[iWidth * iHeight];

	unsigned char* pSourcePixel = pData;
	long* pEnergyDataDest = pEnergyData;
	for(int iY = 0; iY < iHeight; ++iY)
	{
		for(int iX = 0; iX < iWidth; ++iX)
		{
			long iDiffTotal = 0;
			long iDiffCount = 0;
			for(int iYSample = iY - 3; iYSample <= iY + 3; ++iYSample)
			{
				if(iYSample >= 0 && iYSample < iHeight)
				{
					for(int iXSample = iX - 3; iXSample <= iX + 3; ++iXSample)
					{
						if(iXSample >= 0 && iXSample < iWidth && (iXSample != iX || iYSample != iY) )
						{
							for(int iChannel = 0; iChannel < 3; ++iChannel)
							{
								iDiffTotal += abs(pSourcePixel[iChannel] - pData[iYSample * iPitch + (iXSample * 3) + iChannel]);
							}
							++iDiffCount;
						}
					}
				}
			}
			iDiffCount /= 2;
			pEnergyDataDest[0] = (iDiffTotal + iDiffCount / 2) / iDiffCount;

			pSourcePixel += 3;
			pEnergyDataDest += 1;
		}
	}
}

IndexedRetarget::_image_t::_pathfind_t::_pathfind_t()
{
	pCostSrc = 0;
	pIndexDest = 0;

	pParent[0] = 0;
	pParent[1] = 0;
	pParent[2] = 0;
	pChild[0] = 0;
	pChild[1] = 0;
	pChild[2] = 0;
	pShortestChild = 0;

	iCost = 0;
}

void IndexedRetarget::_image_t::CalculateIndexColumns()
{
	delete[] pIndexMap;
	delete[] pIndexVisualisation;
	pIndexVisualisation = 0;
	pIndexMap = new long[iWidth * iHeight];

	m_eIndexOrientation = IO_Columns;

	_pathfind_t* pPathfindingArray = new _pathfind_t[iWidth * iHeight];
	std::vector<_pathfind_t*> vInitialNodes;
	for(int iY = 0; iY < iHeight; ++iY)
	{
		bool bAtTop = iY == 0;
		bool bAtBottom = iY == (iHeight - 1);
		for(int iX = 0; iX < iWidth; ++iX)
		{
			int iOffset = iY * iWidth + iX;
			pPathfindingArray[iOffset].pCostSrc = &pEnergyData[iOffset];
			pPathfindingArray[iOffset].pIndexDest = &pIndexMap[iOffset];

			for(int iXAdjust = (iX == 0 ? 0 : -1); iXAdjust <= 1 && (iX + iXAdjust) < iWidth; ++iXAdjust)
			{
				if(!bAtTop)
					pPathfindingArray[iOffset].pParent[iXAdjust + 1] = &pPathfindingArray[(iY - 1) * iWidth + iX + iXAdjust];
				if(!bAtBottom)
					pPathfindingArray[iOffset].pChild[iXAdjust + 1] = &pPathfindingArray[(iY + 1) * iWidth + iX + iXAdjust];
			}
		}
	}

	for(int iX = 0; iX < iWidth; ++iX)
	{
		pPathfindingArray[iX].iCost = *pPathfindingArray[iX].pCostSrc;
		vInitialNodes.push_back(&pPathfindingArray[iX]);
	}

	DoPathfind(pPathfindingArray, vInitialNodes);

	delete[] pPathfindingArray;
}

void IndexedRetarget::_image_t::CalculateIndexRows()
{
	delete[] pIndexMap;
	delete[] pIndexVisualisation;
	pIndexVisualisation = 0;
	pIndexMap = new long[iWidth * iHeight];

	m_eIndexOrientation = IO_Rows;

	_pathfind_t* pPathfindingArray = new _pathfind_t[iWidth * iHeight];
	std::vector<_pathfind_t*> vInitialNodes;

	for(int iY = 0; iY < iHeight; ++iY)
	{
		for(int iX = 0; iX < iWidth; ++iX)
		{
			bool bAtLeft = iX == 0;
			bool bAtRight = iX == (iWidth - 1);
			int iOffset = iY * iWidth + iX;

			pPathfindingArray[iOffset].pCostSrc = &pEnergyData[iOffset];
			pPathfindingArray[iOffset].pIndexDest = &pIndexMap[iOffset];

			for(int iYAdjust = (iY == 0 ? 0 : -1); iYAdjust <= 1 && (iY + iYAdjust) < iHeight; ++iYAdjust)
			{
				if(!bAtLeft)
					pPathfindingArray[iOffset].pParent[iYAdjust + 1] = &pPathfindingArray[(iY + iYAdjust) * iWidth + iX - 1];
				if(!bAtRight)
					pPathfindingArray[iOffset].pChild[iYAdjust + 1] = &pPathfindingArray[(iY + iYAdjust) * iWidth + iX + 1];
			}
		}

		int iOffset = iY * iWidth;
		pPathfindingArray[iOffset].iCost = *pPathfindingArray[iOffset].pCostSrc;
		vInitialNodes.push_back(&pPathfindingArray[iOffset]);
	}

	DoPathfind(pPathfindingArray, vInitialNodes);

	delete[] pPathfindingArray;
}

void IndexedRetarget::_image_t::DoPathfind(_pathfind_t* pPathfindingArray, std::vector<_pathfind_t*> &vInitialNodes)
{
	long iIndex = 0;
	while(vInitialNodes.size())
	{
		size_t iS = vInitialNodes.size();
		iS = iS;

		for(int iY = 0; iY < iHeight; ++iY)
		{
			for(int iX = 0; iX < iWidth; ++iX)
			{
				pPathfindingArray[iY * iWidth + iX].pShortestChild = 0;
				pPathfindingArray[iY * iWidth + iX].bGotHCost = false;
			}
		}

		std::vector<_pathfind_t*>::iterator itrLowest = vInitialNodes.begin();
		long iLowest = (*itrLowest)->getCost();
		for(std::vector<_pathfind_t*>::iterator itr = vInitialNodes.begin(); itr != vInitialNodes.end(); ++itr)
		{
			if( (*itr)->getCost() < iLowest)
			{
				iLowest = (*itr)->getCost();
				itrLowest = itr;
			}
		}

		_pathfind_t* pPath = *itrLowest;
		vInitialNodes.erase(itrLowest);

		std::stack<_pathfind_t*> stkPathReverse;
		while(pPath)
		{
			stkPathReverse.push(pPath);
			pPath = pPath->pShortestChild;
		}

		while(stkPathReverse.size())
		{
			pPath = stkPathReverse.top();
			stkPathReverse.pop();
			*pPath->pIndexDest = iIndex;

			if(stkPathReverse.size())
			{
				_pathfind_t* LP = pPath->pParent[0]; // Left   parent
				_pathfind_t* CP = pPath->pParent[1]; // Center parent
				_pathfind_t* RP = pPath->pParent[2]; // Right  parent

				if(LP == stkPathReverse.top())
				{
					CP->pChild[1] = LP->pChild[1];
					LP->pChild[1]->pParent[1] = CP;

					if(RP)
						RP->pChild[0] = LP->pChild[1];
					LP->pChild[1]->pParent[2] = RP;

					CP->pChild[0] = LP->pChild[0];
					if(CP->pChild[0])
						CP->pChild[0]->pParent[2] = CP;
				}
				else if(RP == stkPathReverse.top())
				{
					CP->pChild[1] = RP->pChild[1];
					RP->pChild[1]->pParent[1] = CP;

					if(LP)
						LP->pChild[2] = RP->pChild[1];
					RP->pChild[1]->pParent[0] = LP;

					CP->pChild[2] = RP->pChild[2];
					if(CP->pChild[2])
						CP->pChild[2]->pParent[0] = CP;
				}
				else // pBacktrack->pFrom == CP
				{
					if(LP && RP)
					{
						LP->pChild[2] = RP->pChild[1];
						RP->pChild[1]->pParent[0] = LP;

						RP->pChild[0] = LP->pChild[1];
						LP->pChild[1]->pParent[2] = RP;
					}
					else if(LP)
					{
						LP->pChild[2] = 0;
						LP->pChild[1]->pParent[2] = 0;
					}
					else if(RP)
					{
						RP->pChild[0] = 0;
						RP->pChild[1]->pParent[0] = 0;
					}
				}
			}
		}
		++iIndex;
	}
}

bool IndexedRetarget::RetargetDown(int iNewWidth, int iNewHeight)
{
	bool bAdjustRows = iNewHeight != m_pCurrentImage->iHeight;
	bool bAdjustColumns = iNewWidth != m_pCurrentImage->iWidth;

	if(m_pCurrentImage->pParent->m_eIndexOrientation == _image_t::IO_Columns)
	{
		if(bAdjustColumns)
		{
			if(!m_pCurrentImage->adjustColumns(iNewWidth))
				return false;
		}
		if(bAdjustRows)
		{
			bAdjustColumns = false;
			goto reorient_image;
		}
	}
	else
	{
		if(bAdjustRows)
		{
			if(!m_pCurrentImage->adjustRows(iNewHeight))
				return false;
		}
		if(bAdjustColumns)
		{
			bAdjustRows = false;
			goto reorient_image;
		}
	}
	return true;

reorient_image:
	m_pCurrentImage->CalculateEnergyData();
	if(bAdjustRows)
		m_pCurrentImage->CalculateIndexRows();
	else
		m_pCurrentImage->CalculateIndexColumns();
	m_pOriginalImage = m_pCurrentImage;

	m_pCurrentImage = new _image_t;
	m_pCurrentImage->pParent = m_pOriginalImage;
	m_pCurrentImage->iWidth = m_pOriginalImage->iWidth;
	m_pCurrentImage->iPitch = m_pOriginalImage->iWidth * 3;
	m_pCurrentImage->iHeight = m_pOriginalImage->iHeight;
	m_pCurrentImage->pData = new unsigned char[m_pOriginalImage->iPitch * m_pOriginalImage->iHeight];

	if(m_pCurrentImage->pData == 0)
		return false;

	memcpy(m_pCurrentImage->pData, m_pOriginalImage->pData, m_pOriginalImage->iPitch * m_pOriginalImage->iHeight);
	if(bAdjustRows)
		return m_pCurrentImage->adjustRows(iNewHeight);
	else
		return m_pCurrentImage->adjustColumns(iNewWidth);
}

bool IndexedRetarget::_image_t::adjustColumns(int iNewWidth)
{
	if(iNewWidth <= pParent->iWidth) // removing 0 or more seams
	{
		int iIncludeTrip = pParent->iWidth - iNewWidth;
		const unsigned char* pInput = pParent->pData;
		const long* pInputIndex = pParent->pIndexMap;
		unsigned char* pOutput = pData;

		for(int iY = 0; iY < pParent->iHeight; ++iY)
		{
			for(int iX = 0; iX < pParent->iWidth; ++iX)
			{
				if(*pInputIndex >= iIncludeTrip)
				{
					memcpy(pOutput, pInput, 3);
					pOutput += 3;
				}

				++pInputIndex;
				pInput += 3;
			}
		}
	}
	else
	{
		int iNumEachColumn = iNewWidth / pParent->iWidth;

		delete[] pData;
		pData = new unsigned char[iNewWidth * iHeight * 3];
		if(pData == 0)
			return false;

		int iIncludeTrip = iNewWidth - (iNumEachColumn * pParent->iWidth);
		const unsigned char* pInput = pParent->pData;
		const long* pInputIndex = pParent->pIndexMap;
		unsigned char* pOutput = pData;

		for(int iY = 0; iY < pParent->iHeight; ++iY)
		{
			for(int iX = 0; iX < pParent->iWidth; ++iX)
			{
				for(int i = 0; i < iNumEachColumn; ++i)
				{
					memcpy(pOutput, pInput, 3);
					pOutput += 3;
				}
				if(*pInputIndex < iIncludeTrip)
				{
					memcpy(pOutput, pInput, 3);
					pOutput += 3;
				}

				++pInputIndex;
				pInput += 3;
			}
		}
	}
	iWidth = iNewWidth;
	return true;
}

bool IndexedRetarget::_image_t::adjustRows(int iNewHeight)
{
	return true;
}