/* 
	Corsix's Image Retargetr
    Copyright (C) 2007 Peter Cawley

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _INDEXEDRETARGETCLASS_H_
#define _INDEXEDRETARGETCLASS_H_

#include <vector>

class IndexedRetarget
{
public:
	IndexedRetarget();
	~IndexedRetarget();

	bool RetargetDown(int iNewWidth, int iNewHeight);

	bool Init(int iWidth, int iHeight, const unsigned char* pData);

	int GetWidth() const;
	int GetHeight() const;
	const unsigned char* GetImageData() const;
	const unsigned char* GetImageEnergyVisualisation();

protected:
	struct _image_t
	{
		_image_t();
		~_image_t();

		bool adjustRows(int iNewHeight);
		bool adjustColumns(int iNewWidth);
		void visualiseIndex();

		_image_t* pParent;

		long* pIndexMap;
		unsigned char* pIndexVisualisation;

		unsigned char* pData;
		long* pEnergyData;
		int iWidth;
		int iPitch;
		int iHeight;

		enum eIndexOrientations
		{
			IO_Rows,
			IO_Columns
		} m_eIndexOrientation;

		struct _pathfind_t
		{
			_pathfind_t();

			long* pCostSrc;
			long* pIndexDest;

			_pathfind_t *pParent[3];
			_pathfind_t *pChild[3];
			_pathfind_t *pShortestChild;

			long iCost;
			long iHCost;
			bool bGotHCost;

			long getCost();
			long getHCost();
		};
		
		void CalculateEnergyData();
		void CalculateIndexColumns();
		void CalculateIndexRows();

		void DoPathfind(_pathfind_t* pPathfindingArray, std::vector<_pathfind_t*> &vInitialNodes);
	};

	_image_t* m_pCurrentImage;
	_image_t* m_pOriginalImage;
};

#endif