/* 
	Corsix's Image Retargetr
    Copyright (C) 2007 Peter Cawley

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _RETARGETCLASS_H_
#define _RETARGETCLASS_H_

#include <functional>

#include "IndexedRetarget.h"

class Retarget
{
public:
	Retarget();
	~Retarget();

	bool RetargetDown(int iNewWidth, int iNewHeight);

	bool Init(int iWidth, int iHeight, const unsigned char* pData);

	const unsigned char* GetImageData() const;
	const unsigned char* GetImageEnergyVisualisation();

	int GetWidth() const;
	int GetHeight() const;

	void ReCalculateEnergyData();

	void SetPathfindVisCallback(void (*pCallback)(unsigned char*, int, int, void*), void* pToken);
protected:
	struct _image_t
	{
		_image_t();
		_image_t(_image_t* pCopyFrom);
		~_image_t();

		unsigned char* pData;
		long* pEnergyData;
		long* pEnergyDataFlex;
		unsigned char* pEnergyDataVisualised;
		long iAverageEnergy;
		int iWidth;
		int iPitch;
		int iHeight;
		
		void CalculateEnergyData();
	};

	_image_t* m_pCurrentImage;
	_image_t* m_pOriginalImage;

	void (*m_pPathfindCallback)(unsigned char*, int, int, void*);
	void  *m_pPathfindCallbackToken;

	struct _pathfind_t
	{
		long iCost;
		long iHeuristic;
		int iFrom;
		bool bInPath;

		struct less : public std::binary_function<const _pathfind_t*, const _pathfind_t*, bool>
		{
			bool operator()(const _pathfind_t* x, const _pathfind_t* y);
		};
	};

	long _FindColumnToRemove(_pathfind_t* pPathfindArray, int* pResult);
	void _RemoveColumn(int* pColumn);

	long _FindRowToRemove(_pathfind_t* pPathfindArray, int* pResult);
	void _RemoveRow(int* pRow);
};

//typedef IndexedRetarget RetargetrToUse;
typedef Retarget RetargetrToUse;

#endif