/* 
	Corsix's Image Retargetr
    Copyright (C) 2007 Peter Cawley

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <wx/statline.h>
#include "frmMain.h"
#include "frmRealtimeRetarget.h"

BEGIN_EVENT_TABLE(frmMain, wxFrame)
	EVT_BUTTON(IDC_LoadImage, OnLoadImage)
	EVT_BUTTON(IDC_SaveImage, OnSaveImage)
	EVT_CHECKBOX(IDC_ShowEnergy, OnShowEnergyCheck)
	EVT_BUTTON(IDC_Retarget, OnRetarget)
	EVT_BUTTON(IDC_LiveRetarget, OnLiveRetarget)
END_EVENT_TABLE()

BEGIN_EVENT_TABLE(BitmapControl, wxPanel)
	EVT_PAINT(OnPaint)
	//EVT_ERASE_BACKGROUND(OnErase)
END_EVENT_TABLE()

BitmapControl::BitmapControl(wxWindow* pParent, wxBitmap& bmpBitmap)
	: m_bmpBitmap(bmpBitmap)
{
	wxSize oSize(bmpBitmap.GetWidth(), bmpBitmap.GetHeight());
	Create(pParent, wxID_ANY, wxDefaultPosition, oSize);
}

void BitmapControl::SetBitmap(wxBitmap& bmpBitmap)
{
	m_bmpBitmap = bmpBitmap;
}

void BitmapControl::OnPaint(wxPaintEvent& event)
{
	wxPaintDC dcOutput(this);
	wxMemoryDC temp_dc;
	temp_dc.SelectObject(m_bmpBitmap);

	dcOutput.SetClippingRegion(GetSize());
	wxRegionIterator upd(GetUpdateRegion());
	while (upd)
	{
		dcOutput.Blit(upd.GetX(), upd.GetY(), upd.GetW(), upd.GetH(), &temp_dc, upd.GetX(), upd.GetY());
		++upd;
	}
}

void BitmapControl::OnErase(wxEraseEvent& event)
{
}

frmMain::frmMain(const wxString& title, const wxPoint& pos, const wxSize& size)
	: wxFrame((wxFrame *)NULL, wxID_ANY, title, pos, size, wxDEFAULT_FRAME_STYLE)
	, m_bmpImage(8, 8)
	, m_dlgImageChooser(this, wxT("Choose image file"), wxT(""), wxT(""), wxT("Supported Image Formats (*.bmp;*.gif;*.jpg;*.jpeg;*.pcx;*.png;*.tif;*.tiff;*.xpm;*.ico;*.cur)|*.bmp;*.gif;*.jpg;*.jpeg;*.pcx;*.png;*.tif;*.tiff;*.xpm;*.ico;*.cur|All Files (*.*)|*.*"))
{
	wxBoxSizer *pMainSizer = new wxBoxSizer( wxHORIZONTAL );
	wxBoxSizer *pCommandsSizer = new wxBoxSizer( wxVERTICAL );

	m_oRetargetr.Init(8, 8, 0);

	// Preview box
	wxStaticBoxSizer* pPreviewSizer = new wxStaticBoxSizer( wxHORIZONTAL, this, wxT("Image Preview") );
	pPreviewSizer->Add(m_pImagePreview = new BitmapControl(this, m_bmpImage), 1, wxALL | wxFIXED_MINSIZE | wxEXPAND, 2);
	pMainSizer->Add(pPreviewSizer, 1, wxEXPAND);

	// Image box
	wxStaticBoxSizer *pImageContainer = new wxStaticBoxSizer( wxVERTICAL, this, wxT("Image") );
	pImageContainer->Add(new wxButton(this, IDC_LoadImage, wxT("Load")), 0, wxALIGN_CENTER | wxALL, 2);
	pImageContainer->Add(new wxButton(this, IDC_SaveImage, wxT("Save")), 0, wxALIGN_CENTER | wxALL, 2);
	pImageContainer->Add(m_chkShowEnergy = new wxCheckBox(this, IDC_ShowEnergy, wxT("Show Energy")),  0, wxALIGN_CENTER | wxALL, 2);

	pCommandsSizer->Add(pImageContainer, 0, wxALIGN_LEFT | wxEXPAND);

	// Retarget box
	wxStaticBoxSizer *pRetargetContainer = new wxStaticBoxSizer( wxVERTICAL, this, wxT("Retarget") );

	wxFlexGridSizer *pRetargetSettings = new wxFlexGridSizer(2);
	pRetargetSettings->SetFlexibleDirection(wxHORIZONTAL);
	pRetargetSettings->AddGrowableCol(1, 1);

	pRetargetSettings->Add(new wxStaticText(this, wxID_ANY, wxT("New Width")), 0, wxALIGN_CENTER_VERTICAL | wxFIXED_MINSIZE | wxALL, 2);
	pRetargetSettings->Add(m_txtRetargetWidth = new wxTextCtrl(this, wxID_ANY, wxT("256")), 1, wxEXPAND | wxALIGN_CENTER_VERTICAL | wxALL, 2);

	pRetargetSettings->Add(new wxStaticText(this, wxID_ANY, wxT("New Height")), 0, wxALIGN_CENTER_VERTICAL | wxFIXED_MINSIZE | wxALL, 2);
	pRetargetSettings->Add(m_txtRetargetHeight = new wxTextCtrl(this, wxID_ANY, wxT("256")), 1, wxEXPAND | wxALIGN_CENTER_VERTICAL | wxALL, 2);

	pRetargetContainer->Add(pRetargetSettings, 0, wxEXPAND | wxALIGN_LEFT);
	pRetargetContainer->Add(m_chkShowPathing = new wxCheckBox(this, wxID_ANY, wxT("Show Pathfinding")), 0, wxEXPAND | wxALIGN_LEFT);
	pRetargetContainer->Add(m_btnRetarget = new wxButton(this, IDC_Retarget, wxT("Retarget")), 0, wxALIGN_CENTER | wxALL, 2);
	pRetargetContainer->Add(new wxStaticLine(this), 0, wxEXPAND | wxALL, 1);
	pRetargetContainer->Add(new wxButton(this, IDC_LiveRetarget, wxT("Live Retarget")), 0, wxALIGN_CENTER | wxALL, 2);

	pCommandsSizer->Add(pRetargetContainer, 0, wxALIGN_LEFT | wxEXPAND);

	// Finish up
	pMainSizer->Add(pCommandsSizer, 0, wxALIGN_LEFT | wxLEFT, 2);

	SetSizer(pMainSizer);
	pMainSizer->SetSizeHints( this );

	SetBackgroundColour(m_btnRetarget->GetBackgroundColour());
}

void frmMain::OnShowEnergyCheck(wxCommandEvent& e)
{
	const unsigned char* pData = e.IsChecked() ? m_oRetargetr.GetImageEnergyVisualisation() : m_oRetargetr.GetImageData();
	wxImage oImage(m_bmpImage.GetWidth(), m_bmpImage.GetHeight(), const_cast<unsigned char*>(pData), true);
	m_bmpImage = wxBitmap(oImage);

	Refresh();
}

void frmMain::OnLoadImage(wxCommandEvent& e)
{
	m_dlgImageChooser.SetWindowStyleFlag(wxFD_OPEN | wxFD_FILE_MUST_EXIST | wxFD_PREVIEW);
	if(m_dlgImageChooser.ShowModal() == wxID_OK)
	{
		wxImage oImage;
		if(oImage.LoadFile(m_dlgImageChooser.GetPath()))
		{
			m_bmpImage = wxBitmap(oImage);
			m_pImagePreview->SetBitmap(m_bmpImage);
			m_oRetargetr.Init(oImage.GetWidth(), oImage.GetHeight(), oImage.GetData());

			wxString sWidth, sHeight;
			sWidth.Printf(wxT("%i"), (int)oImage.GetWidth());
			m_txtRetargetWidth->SetValue(sWidth);

			sHeight.Printf(wxT("%i"), (int)oImage.GetHeight());
			m_txtRetargetHeight->SetValue(sHeight);

			m_chkShowEnergy->SetValue(false);

			Refresh();
		}
		else
		{
			::wxMessageBox(wxT("Unable to load file"), wxT("Load Image"), wxICON_ERROR, this);
		}
	}
}

void frmMain::PathfindVisCallback(unsigned char* pData, int iWidth, int iHeight, frmMain* pForm)
{
	if( ++pForm->m_iPathfindVisTimer == 128)
	{
		bool bNewSize = iWidth < pForm->m_bmpImage.GetWidth() || iHeight < pForm->m_bmpImage.GetHeight();

		wxImage oImage(iWidth, iHeight, const_cast<unsigned char*>(pData), true);
		pForm->m_bmpImage = wxBitmap(oImage);

		pForm->m_pImagePreview->SetBitmap(pForm->m_bmpImage);
		pForm->m_pImagePreview->Refresh(bNewSize);
		wxSafeYield();
		pForm->m_iPathfindVisTimer = 0;
	}
}

void frmMain::OnRetarget(wxCommandEvent& e)
{
	long iNewWidth, iNewHeight;
	if(!m_txtRetargetWidth->GetValue().ToLong(&iNewWidth) || !m_txtRetargetHeight->GetValue().ToLong(&iNewHeight))
	{
		::wxMessageBox(wxT("Width/Height must be numbers"), wxT("Retarget"), wxICON_ERROR, this);
		return;
	}

	/*
	m_iPathfindVisTimer = 0;
	if(m_chkShowPathing->IsChecked())
		m_oRetargetr.SetPathfindVisCallback( (void(*)(unsigned char*,int,int,void*)) PathfindVisCallback, this);
	else
		m_oRetargetr.SetPathfindVisCallback(0, 0);
	*/

	if(!m_oRetargetr.RetargetDown(iNewWidth, iNewHeight))
	{
		::wxMessageBox(wxT("Retargetting Error"), wxT("Retarget"), wxICON_ERROR, this);
		return;
	}

	const unsigned char* pData = m_chkShowEnergy->IsChecked() ? m_oRetargetr.GetImageEnergyVisualisation() : m_oRetargetr.GetImageData();
	wxImage oImage(iNewWidth, iNewHeight, const_cast<unsigned char*>(pData), true);
	m_bmpImage = wxBitmap(oImage);

	Refresh();
}

void frmMain::OnSaveImage(wxCommandEvent& e)
{
}

void frmMain::OnLiveRetarget(wxCommandEvent& e)
{
	frmRealtimeRetarget frmLive(this, wxT("Resize window to retarget image"), wxDefaultPosition, wxSize(m_bmpImage.GetWidth(), m_bmpImage.GetHeight()), m_oRetargetr);
	frmLive.ShowModal();
}